from __future__ import annotations
from lalalang.evaluator.object import Object


class Environment:
    """
    The environment is just a table with the list of identifiers
    and its values associated, it has also a outer environment since
    we can access global values
    """

    @staticmethod
    def empty():
        return Environment(dict({}), None)

    def __init__(self, store: dict[str, Object], outer: Environment):
        self.store: dict[str, Object] = store
        self.outer: Environment = outer

    def get_local(self, name: str) -> Object:
        value: Object = self.store.get(name)
        if not value and self.outer != None:
            value = self.outer.get_local(name)
        return value

    def set_local(self, name: str, val: Object) -> Object:
        self.store[name] = val
        return val



from lalalang.parser.ast import *
from lalalang.evaluator.object import *
from lalalang.evaluator.environment import Environment

# References
TRUE: Boolean = Boolean(True)
FALSE: Boolean = Boolean(False)
NULL: Null = Null()

from lalalang.evaluator.builtins import BUILTINS


def eval_3lang(node: Node, env: Environment) -> Object:
    """
    This is executing a tree walk interpreter, doing a kind of
    postorder traverse over the ast
    """

    # Statements
    if isinstance(node, Program):
        return eval_program(node, env)

    elif isinstance(node, ExpressionStatement):
        return eval_3lang(node.expression, env)

    elif isinstance(node, BlockStatement):
        return eval_block_statement(node, env)

    elif isinstance(node, ReturnStatement):
        value = eval_3lang(node.return_value, env)
        if is_error(value):
            return value
        return ReturnValue(value)

    elif isinstance(node, LetStatement):
        value = eval_3lang(node.value, env)
        if is_error(value):
            return value
        return env.set_local(node.name.value, value)

    # Expressions
    elif isinstance(node, PrefixExpression):
        right = eval_3lang(node.right, env)
        if is_error(right):
            return right
        return eval_prefix_expression(node.operator, right)

    elif isinstance(node, InfixExpression):
        left = eval_3lang(node.left, env)
        if is_error(left):
            return left

        right = eval_3lang(node.right, env)
        if is_error(right):
            return right

        return eval_infix_expression(node.operator, left, right)

    elif isinstance(node, IfExpression):
        condition = eval_3lang(node.condition, env)
        if is_error(condition):
            return condition
        return eval_if_expression(node, env)

    elif isinstance(node, Identifier):
        return eval_identifier(node, env)

    elif isinstance(node, CallExpression):
        fun = eval_3lang(node.function, env)
        if is_error(fun):
            return fun

        args = eval_expressions(node.arguments, env)
        if len(args) == 1 and is_error(args[0]):
            return args[0]

        return apply_function(fun, args)

    # Internal objects
    elif isinstance(node, IntegerLiteral):
        return Integer(node.value)

    elif isinstance(node, BooleanLiteral):
        return boolean_reference(node.value)

    elif isinstance(node, StringLiteral):
        return String(node.value)

    elif isinstance(node, NullLiteral):
        return NULL

    elif isinstance(node, FunctionLiteral):
        return Function(node.parameters, node.body, env)

    return None


def eval_program(root: Program, env: Environment) -> Object:
    """
    The root of every program is a list of statements, not
    the ast nodes, we need to start from the root
    """
    result: Object = None

    for statement in root.statements:
        result = eval_3lang(statement, env)

        if isinstance(result, ReturnValue):
            return result.value

        elif isinstance(result, Error):
            return result

    return result


def eval_block_statement(block: BlockStatement, env: Environment) -> Object:
    """
    This is somewhat similar to eval_program except that is
    looking for any value to return and stop in the statement
    """
    result: Object = None

    for statement in block.statements:

        if result := eval_3lang(statement, env):
            rt: ObjectType = result.object_type()

            if rt == ObjectType.RETURN_VALUE or rt == ObjectType.ERROR:
                return result

    return result


def eval_prefix_expression(operator: str, right: Object) -> Object:
    """We support evaluating the two only prefix operators"""
    if operator == "!":
        return eval_bang_operator_expression(right)
    elif operator == "-":
        return eval_minus_operator_expression(right)
    return Error("Unknown operator %s %s" % (operator, right.object_type()))


def eval_bang_operator_expression(right: Object) -> Object:
    """
    If the object is null or false, it should evaluate to true, any
    othe expression or object should be false
    """
    if right in [FALSE, NULL]:
        return TRUE
    return FALSE


def eval_minus_operator_expression(right: Object) -> Object:
    """Just reverse the value is already in the object"""
    if isinstance(right, Integer):
        return Integer(-right.value)

    return Error("Unknown operator -%s" % right.object_type())


def eval_infix_expression(operator: str, left: Object, right: Object) -> Object:
    """Main method for evaluating infix expression"""
    if left.object_type() != right.object_type():
        return Error(
            "Type missmatch %s %s %s"
            % (left.object_type(), operator, right.object_type())
        )

    result: Object = None

    if isinstance(left, Integer) and isinstance(right, Integer):
        result = eval_integer_infix_expression(operator, left, right)
    elif isinstance(left, Boolean) and isinstance(right, Boolean):
        result = eval_boolean_infix_expression(operator, left, right)
    elif isinstance(left, String) and isinstance(right, String):
        result = eval_string_infix_expression(operator, left, right)

    if not result:
        return Error(
            "Unkown operator %s %s %s"
            % (left.object_type(), operator, right.object_type())
        )

    return result


def eval_integer_infix_expression(
    operator: str, left: Integer, right: Integer
) -> Object:
    """
    Here we evaulate relational and arithmetic operations
    with integers
    """

    # Arithmetic
    if operator == "+":
        return Integer(left.value + right.value)
    elif operator == "-":
        return Integer(left.value - right.value)
    elif operator == "*":
        return Integer(left.value * right.value)
    elif operator == "/":
        return Integer(left.value // right.value)
    elif operator == "%":
        return Integer(left.value % right.value)

    # Relational
    elif operator == "<":
        return boolean_reference(left.value < right.value)
    elif operator == ">":
        return boolean_reference(left.value > right.value)
    elif operator == "==":
        return boolean_reference(left.value == right.value)
    elif operator == "!=":
        return boolean_reference(left.value != right.value)

    return None


def eval_boolean_infix_expression(
    operator: str, left: Boolean, right: Boolean
) -> Object:
    """Here we evaluate logical expressions with booleans"""
    if operator == "&&":
        return boolean_reference(left.value and right.value)
    elif operator == "||":
        return boolean_reference(left.value or left.value)
    return None


def eval_string_infix_expression(operator: str, left: String, right: String) -> Object:
    """Here we are adding a way for concatenating strings"""
    if operator == "+":
        return String(left.value + right.value)
    return None


def eval_if_expression(expression: IfExpression, env: Environment) -> Object:
    """
    Here we parse the condition and depending on the result we
    execute the consequence or alternative
    """
    predicate: Object = eval_3lang(expression.condition, env)

    if is_truthy(predicate):
        return eval_3lang(expression.consequence, env)
    elif expression.alternative is not None:
        return eval_3lang(expression.alternative, env)

    return NULL


def is_truthy(obj: Object) -> bool:
    """Here we check which kind of objects evaluate to true or false"""
    if obj in [NULL, FALSE]:
        return False
    return True


def eval_identifier(node: Identifier, env: Environment) -> Object:
    """Check if the name has a value associated in the environment"""
    if value := env.get_local(node.value):
        return value

    if proc := BUILTINS.get(node.value):
        return proc

    return Error("Identifier not found: %s" % node.value)


def eval_expressions(expressions: list[Expression], env: Environment) -> list[Object]:
    """Here we evaluate a bunch of expressions and keep track of them"""
    result: list[Object] = []
    for exp in expressions:
        evaluated = eval_3lang(exp, env)
        if is_error(evaluated):
            return [evaluated]
        result.append(evaluated)
    return result


def boolean_reference(value: bool) -> Boolean:
    """Return the Boolean reference instead of creating new value"""
    if value:
        return TRUE
    return FALSE


def apply_function(function: Object, args: list[Object]) -> Object:
    """
    We get the current environment and the outer one to run the
    body of the function with its parameters
    """
    if isinstance(function, Function):
        extended_env: Environment = extend_function_env(function, args)
        evaluated: Object = eval_3lang(function.body, extended_env)

        if isinstance(evaluated, ReturnValue):
            return evaluated.value
        return evaluated

    elif isinstance(function, Builtin):
        return function.fun(*args)

    return Error("Not a function: %s" % function.object_type())


def extend_function_env(function: Function, args: list[Object]) -> Environment:
    """Add the outer env to the local one"""
    env: Environment = Environment(dict({}), function.env)

    for index, param in enumerate(function.parameters):
        env.set_local(param.value, args[index])

    return env


def is_error(obj: Object) -> bool:
    """Helper method to check whenever an exception occurs"""
    if obj:
        return obj.object_type() == ObjectType.ERROR
    return False



from unittest import TestCase
from tests.mocks.evaluator import *
from lalalang.lexer.lexer import Lexer
from lalalang.parser.parser import Parser
from lalalang.evaluator.evaluator import eval_3lang


class TestEvaluator(TestCase):
    def test_integer(self):
        self.check_evaluation(INTEGER)

    def test_boolean(self):
        self.check_evaluation(BOOLEAN)

    def test_prefix(self):
        self.check_evaluation(PREFIX)

    def test_infix(self):
        self.check_evaluation(INFIX)

    def test_conditionals(self):
        self.check_evaluation(CONDITIONALS)

    def test_return(self):
        self.check_evaluation(RETURN)

    def test_errors(self):
        self.check_evaluation(ERRORS)

    def test_function_calls(self):
        self.check_evaluation(FUNCTION_CALLS)

    def test_environments(self):
        self.check_evaluation(ENVIRONMENT)

    def test_builtins(self):
        self.check_evaluation(BUILTINS)

    def test_logical(self):
        self.check_evaluation(LOGICAL)

    def test_null(self):
        self.check_evaluation(NULLS)

    def test_strings(self):
        self.check_evaluation(STRINGS)

    def check_evaluation(self, cases):
        for case in cases:
            source = case.get("input")
            internal = self.build_object(source)
            expected = case.get("expected")
            self.assertEqual(internal.inspect(), expected)

    def build_object(self, source):
        lex = Lexer(source)
        par = Parser(lex)
        program = par.parse_program()
        env = Environment.empty()
        return eval_3lang(program, env)



from abc import ABC, abstractmethod


class Node(ABC):
    """
    This class represents every node that is part of the
    abstract syntax tree
    """

    @abstractmethod
    def token_literal(self) -> str:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class Statement(Node):
    """
    These nodes are reserved only for the statements, meaning
    it won't produce any value. In la la lang there there are
    only two staments return and let.
    """

    @abstractmethod
    def statement_node(self) -> None:
        pass


class Expression(Node):
    """
    These nodes are reserved only for the expression, meaning it
    will produce a value. Aside from let and return everything is
    a expression.
    """

    @abstractmethod
    def expression_node(self) -> None:
        pass



from __future__ import annotations
from enum import Enum
from typing import Callable
from dataclasses import dataclass
from abc import ABC, abstractmethod
from lalalang.parser.ast import Identifier, BlockStatement


class Object(ABC):
    """
    This is how we internally represent the objects in the
    interpreter in order to manipulate it
    """

    @abstractmethod
    def object_type(self) -> ObjectType:
        pass

    @abstractmethod
    def inspect(self) -> str:
        pass


from lalalang.evaluator.environment import Environment  # Avoiding circular import


class ObjectType(Enum):
    """These are all the possible objects we can have"""

    INTEGER = "INTEGER"
    BOOLEAN = "BOOLEAN"
    NULL = "NULL"
    RETURN_VALUE = "RETURN_VALUE"
    ERROR = "ERROR"
    FUNCTION = "FUNCTION"
    BUILTIN = "BUILTIN"
    STRING = "STRING"


@dataclass
class Integer(Object):
    def __init__(self, value: int):
        self.value: int = value

    def object_type(self) -> ObjectType:
        return ObjectType.INTEGER

    def inspect(self) -> str:
        return str(self.value)


class String(Object):
    def __init__(self, value: str):
        self.value: str = value

    def object_type(self) -> ObjectType:
        return ObjectType.STRING

    def inspect(self) -> str:
        return self.value


class Boolean(Object):
    def __init__(self, value: bool):
        self.value: bool = value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Boolean):
            return self.value == other.value
        return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Boolean):
            return self.value != other.value
        return False

    def object_type(self) -> ObjectType:
        return ObjectType.BOOLEAN

    def inspect(self) -> str:
        return str(self.value).lower()


class Null(Object):
    def object_type(self) -> ObjectType:
        return ObjectType.NULL

    def inspect(self) -> str:
        return "null"


class ReturnValue(Object):
    def __init__(self, value: Object):
        self.value: Object = value

    def object_type(self) -> ObjectType:
        return ObjectType.RETURN_VALUE

    def inspect(self) -> str:
        return self.value.inspect()


class Error(Object):
    def __init__(self, message: str):
        self.message: str = message

    def object_type(self) -> ObjectType:
        return ObjectType.ERROR

    def inspect(self) -> str:
        return "ERROR: %s" % self.message


class Function(Object):
    def __init__(
        self, parameters: list[Identifier], body: BlockStatement, env: Environment
    ):
        self.env: Environment = env
        self.body: BlockStatement = body
        self.parameters: list[Identifier] = parameters

    def object_type(self) -> ObjectType:
        return ObjectType.FUNCTION

    def inspect(self) -> str:
        params: list[str] = [str(i) for i in self.parameters]
        return "fun(%s){\n\t%s\n}" % (",".join(params), self.body)


class Builtin(Object):
    def __init__(self, fun: Callable):
        self.fun: Callable = fun

    def object_type(self) -> ObjectType:
        return ObjectType.BUILTIN

    def inspect(self) -> str:
        return "Builtin function"



from lalalang.evaluator.evaluator import *
from lalalang.evaluator.object import *


INTEGER = [
    {
        "input": "314",
        "expected": "314",
    },
    {
        "input": "016",
        "expected": "16",
    },
]

BOOLEAN = [
    {
        "input": "true",
        "expected": "true",
    },
    {
        "input": "false",
        "expected": "false",
    },
]

PREFIX = [
    {
        "input": "-5",
        "expected": "-5",
    },
    {
        "input": "--5",
        "expected": "5",
    },
    {
        "input": "!true",
        "expected": "false",
    },
    {
        "input": "!false",
        "expected": "true",
    },
]

INFIX = [
    {
        "input": "5 + 4",
        "expected": "9",
    },
    {
        "input": "13 * 2 + 6",
        "expected": "32",
    },
    {
        "input": "(5 + 2) * 3",
        "expected": "21",
    },
    {
        "input": "(3 / 3) * 3",
        "expected": "3",
    },
    {
        "input": "10 - (4 + 2)",
        "expected": "4",
    },
    {
        "input": "10 - 20",
        "expected": "-10",
    },
    {
        "input": "10 % 2",
        "expected": "0",
    },
]

LOGICAL = [
    {
        "input": "true || false",
        "expected": "true",
    },
    {
        "input": "false || false",
        "expected": "false",
    },
    {
        "input": "false && true",
        "expected": "false",
    },
    {
        "input": "true && true",
        "expected": "true",
    },
]

CONDITIONALS = [
    {
        "input": "if(5 == 5) { 3 }",
        "expected": "3",
    },
    {
        "input": "if(10 < 3) { 4 } else { 5 }",
        "expected": "5",
    },
    {
        "input": "if(5 != 5) { 3 }",
        "expected": "null",
    },
]

RETURN = [
    {
        "input": "let a = 5*8;return 10;",
        "expected": "10",
    },
    {
        "input": "return 2; return 3;",
        "expected": "2",
    },
    {
        "input": "if(1 == 1){ if(1 == 1) { return 1; } return 0; }",
        "expected": "1",
    },
]

ERRORS = [
    {
        "input": "-true",
        "expected": "ERROR: Unknown operator -ObjectType.BOOLEAN",
    },
    {
        "input": "10 + false",
        "expected": "ERROR: Type missmatch ObjectType.INTEGER + ObjectType.BOOLEAN",
    },
    {
        "input": "false + false",
        "expected": "ERROR: Unkown operator ObjectType.BOOLEAN + ObjectType.BOOLEAN",
    },
    {
        "input": "a",
        "expected": "ERROR: Identifier not found: a",
    },
    {
        "input": "return x",
        "expected": "ERROR: Identifier not found: x",
    },
]

FUNCTION_CALLS = [
    {
        "input": "let a = fun(x) { x + 1 }; a(2)",
        "expected": "3",
    },
    {
        "input": "let b = fun() { false }; b()",
        "expected": "false",
    },
    {
        "input": "let c = fun(x) { if(x > 0) { return 1; } return 2; }; c(1)",
        "expected": "1",
    },
]

ENVIRONMENT = [
    {
        "input": "let a = 10; let b = 5; a + b;",
        "expected": "15",
    },
    {
        "input": "let a = 10; let f = fun() { return a; }; f()",
        "expected": "10",
    },
    {
        "input": "let x = 10; let f = fun() { let y = 3; return x + y; }; f()",
        "expected": "13",
    },
]

BUILTINS = [
    {
        "input": "print(10)",
        "expected": "null",
    },
    {
        "input": "println(10)",
        "expected": "null",
    },
    {
        "input": '"Universe: " + toString(10)',
        "expected": "Universe: 10",
    },
]

NULLS = [
    {
        "input": "null",
        "expected": "null",
    },
    {
        "input": "let a = null; a;",
        "expected": "null",
    },
    {
        "input": "let a = if(false) { 1 }; a;",
        "expected": "null",
    },
]

STRINGS = [
    {
        "input": '"Feliz jueves!"',
        "expected": "Feliz jueves!",
    },
    {
        "input": 'let a = "b"; a;',
        "expected": "b",
    },
    {
        "input": 'let b = fun() { return "b"; }; b()',
        "expected": "b",
    },
    {
        "input": '"a" + "b"',
        "expected": "ab",
    },
    {
        "input": '"a" + "b" + "c"',
        "expected": "abc",
    },
    {
        "input": '"feliz" + " " + "jueves"',
        "expected": "feliz jueves",
    },
]



from lalalang.evaluator.evaluator import NULL
from lalalang.evaluator.object import *


def _print_line(*args: Object) -> Object:
    """Internal representation of println function"""
    if len(args) != 1:
        return Error("It must recieve one parameter")

    print(args[0].inspect())
    return NULL


def _print(*args: Object) -> Object:
    """Internal representation of print function"""
    if len(args) != 1:
        return Error("It must recieve one parameter")

    print(args[0].inspect(), end="")
    return NULL


def _to_string(*args: Object) -> Object:
    """Convert an integer to a string"""
    if len(args) != 1:
        return Error("It must recieve one parameter")

    number: Object = args[0]

    if not isinstance(number, Integer):
        return Error("It must be an integer")

    return String(str(number.value))


BUILTINS: dict[str, Object] = {
    "println": Builtin(_print_line),
    "print": Builtin(_print),
    "toString": Builtin(_to_string),
}



import sys
import click
import signal
from functools import wraps
from lalalang.parser.parser import Parser
from lalalang.lexer.lexer import Lexer
from lalalang.lexer.token import TokenType, Token
from lalalang.evaluator.object import Null
from lalalang.evaluator.evaluator import eval_3lang
from lalalang.evaluator.environment import Environment


@click.command()
@click.option("--mode", default="eval", help="REPL mode (lex|parse|eval)")
@click.option("--src", default=None, help="Input file with 3lang code")
def cli(mode, src):
    code = None
    if src:
        with open(src, "r") as f:
            code = f.read()
    else:
        print("Welcome to the city of stars!🌟")
        print("This is the La La Lang Programming Language v1.2!")

    if mode == "lex":
        lexing(code)

    elif mode == "parse":
        parsing(code)

    elif mode == "eval":
        env = Environment.empty()
        evaluating(code, env=env)


def read_eval_print_loop(func):
    @wraps(func)
    def inner_function(code, **kwargs):
        if not code:
            while True:
                line = input("♪♪ > ")
                if line == "":
                    continue
                func(line, **kwargs)
        func(code, **kwargs)

    return inner_function


@read_eval_print_loop
def lexing(code):
    lex = Lexer(code)
    token = Token.empty()
    while token.token_type != TokenType.EOF:
        token = lex.next_token()
        print(repr(token))


@read_eval_print_loop
def parsing(code):
    lex = Lexer(code)
    par = Parser(lex)
    program = par.parse_program()

    if len(par.errors) == 0:
        print(program)

    [print(err) for err in par.errors]


@read_eval_print_loop
def evaluating(code, env):
    lex = Lexer(code)
    par = Parser(lex)
    program = par.parse_program()

    if len(par.errors) > 0:
        [print(i) for i in par.errors]
        return

    evaluated = eval_3lang(program, env)
    if evaluated and not isinstance(evaluated, Null):
        print(evaluated.inspect())


def ctrlc_handler(sig, fram):
    print("\nSee you at Seb's 😉")
    sys.exit(0)


signal.signal(signal.SIGINT, ctrlc_handler)


if __name__ == "__main__":
    cli()
